# -*- coding: utf-8 -*-
import json
import logging
from typing import List, Set, Optional
from ..client.redis_client import Redis
from ..lib import cfg
from ..model.items import AuthUser

logger = logging.getLogger(__name__)
rds = Redis(redis_uri=cfg.get('AUTH_REDIS_URL', None), redis_db=cfg.get('AUTH_REDIS_DB', None)).client
default_expire = 43200


def save_cached_user(access_token: str, user: AuthUser, expire: int = None):
    if user:
        _expire = expire or default_expire
        rds.set(f'share:access:{access_token}', user.json(), ex=_expire)


def get_cached_user(access_token: str) -> Optional[AuthUser]:
    rs = rds.get(f'share:access:{access_token}')
    if rs:
        js_user = json.loads(rs)
        if isinstance(js_user, str):
            js_user = json.loads(js_user)
        user = AuthUser(**js_user)
        user.userName = js_user.get('username', '')
        return user

    return None


def has_user(access_token: str) -> bool:
    return rds.exists(f'share:access:{access_token}')


def save_cached_permission(access_token: str, permissions: Set[str], expire: int = None):
    if permissions:
        key = f'share:permissions:{access_token}'
        for per in permissions:
            rds.sadd(key, per)
        _expire = expire or default_expire
        rds.expire(key, _expire)


def has_permission(access_token: str, permission: str) -> bool:
    return rds.sismember(f'share:permissions:{access_token}', f'"{permission}"')


def get_cached_permission(access_token: str) -> List[str]:
    return rds.smembers(f'share:permissions:{access_token}')


def clean_cache(access_token: str):
    rds.delete(f'share:access:{access_token}')
    rds.delete(f'share:permissions:{access_token}')


def get_user_by_token(token_str: str) -> Optional[AuthUser]:
    rs = rds.get(f'share:access:{token_str}')
    if rs:
        js_user = json.loads(rs)
        if isinstance(js_user, str):
            js_user = json.loads(js_user)
        user = AuthUser(**js_user)
        user.userName = js_user.get('username', '') or js_user.get('name', '')
        return user

    return None



from enum import IntEnum
from typing import Dict, Any, List

from ..model.base import ExecRet


class ErrCode(IntEnum):
    """
    error codes
    """
    SUCCESS = 0
    ERROR = -1


class Resp(ExecRet):
    """
    response body
    """
    code: IntEnum = ErrCode.SUCCESS

    @classmethod
    def ok(cls, data: Any = None, msg: str = '操作成功') -> Dict[str, Any]:
        """
        generate success response body
        :param data: data
        :param msg: msg
        :return: json dict resp body
        """
        return cls(
            success=True,
            data=data,
            msg=msg,
            code=ErrCode.SUCCESS
        ).dict()

    @classmethod
    def err(cls, data: Any = None, msg: str = '操作失败', code: IntEnum = ErrCode.ERROR) -> Dict[str, Any]:
        """
        generate error response body with external status code 200
        :param data: data
        :param msg: msg
        :param code: user defined (not http) status code
        :return: json dict resp body
        """
        if code == ErrCode.SUCCESS:
            code = ErrCode.ERROR
        return cls(
            success=False,
            data=data,
            msg=msg,
            code=code
        ).dict()

    @classmethod
    def page(cls, rows: List = [], total_count: int = 0, msg: str = '操作成功') -> \
            Dict[
                str, Any]:
        """
        generate page response body
        :param rows:
        :param total:
        :param msg:
        :return:
        """
        return cls(
            success=True,
            data={'list': rows, 'total': total_count},
            msg=msg,
            code=ErrCode.SUCCESS
        ).dict()

    @staticmethod
    def is_success(result: Dict) -> bool:
        """
        判断响应是否为成功
        :param result:
        :return:
        """
        if not result or not result.get('success', False):
            return False

        return True

    @staticmethod
    def get_data(result: dict, force: bool = False) -> object:
        """
        判断响应是否为成功
        :param result: 响应结果
        :param force: 是否跳过响应状态验证
        :return:
        """
        if not force:
            if not Resp.is_success(result):
                return None

        return result.get('data', None)



# -*- coding: utf-8 -*-
import oss2
from typing import Optional, Dict
from ..lib import cfg, logger
from ..exception.exceptions import NoConfigException

LOGGER = logger.get('Oss2Client')


# oss2客户端简易封装
class Oss2Client(object):
    def __init__(self, endpoint: str = None, access_key: str = None, secret_key: str = None):
        self.access_key = access_key or cfg.get("OSS2_ACCESS_KEY")
        self.secret_key = secret_key or cfg.get("OSS2_SECRET_KEY")
        self.endpoint = endpoint or cfg.get("OSS2_ENDPOINT") or "https://oss-cn-beijing.aliyuncs.com"
        # 文件下载链接：示例 https://%s.oss-cn-beijing.aliyuncs.com
        self.share_url = cfg.get('OSS2_SHARE_URL_FORMAT') or self.endpoint.replace('://', '://%s.').replace('-internal.', '.')
        if not self.access_key:
            raise NoConfigException("oss2 access key not config!")
        if not self.secret_key:
            raise NoConfigException("oss2 secret key not config!")

        self.auth = oss2.Auth(self.access_key, self.secret_key)
        self.bucket_map = dict()

    def put_stream(self, bucket_name: str, object_name: str, image_content: bytes, content_type='application/octet-stream') -> Dict:
        bucket = self._get_bucket(bucket_name)
        headers = dict()
        if content_type:
            headers["Content-Type"] = content_type
        if object_name:
            bucket.put_object(key=object_name, data=image_content, headers=headers)
            file_url = self.get_object_url(bucket_name, object_name)
            return {
                "url": file_url,
                "file_path": f"{bucket_name}/{object_name}"
            }

    def put_object(self, bucket_name: str, object_name: str, file_path: str, content_type='application/octet-stream') -> Dict:
        bucket = self._get_bucket(bucket_name)
        headers = dict()
        if content_type:
            headers["Content-Type"] = content_type
        if object_name:
            bucket.put_object_from_file(filename=file_path, key=object_name, headers=headers)
            file_url = self.get_object_url(bucket_name, object_name)
            return {
                "url": file_url,
                "file_path": object_name
            }

    def _get_bucket(self, bucket_name: str):
        bucket = self.bucket_map.get(bucket_name)
        if not bucket:
            bucket = oss2.Bucket(self.auth, self.endpoint, bucket_name)
            self.bucket_map[bucket_name] = bucket

        return bucket

    def get_object_url(self, bucket_name: str, object_name: str) -> Optional[str]:
        # 桶均为公共读私有写，可直接访问对象
        if '%s' in self.share_url:
            return f'{self.share_url % bucket_name}/{object_name}'

        # 自有域名
        return f'{self.share_url}/{object_name}'



import uuid
import traceback
from typing import List

from starlette.requests import Request
from starlette.responses import StreamingResponse, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response
from ..auth.share_auth import ShareAuth
from ..exception.exceptions import NoAuthException
from ..lib import logger
from ..model.handler import ErrCode

# 固定白名单
WHITELIST = [
]

LOGGER = logger.get('鉴权中间件')
share_auth = ShareAuth()


class AuthMiddleware(BaseHTTPMiddleware):
    """ 鉴权中间件 """

    def __init__(self, app, whitelist: List[str] = None):
        self.FASTAPI_APP = app.app.app
        if whitelist:
            WHITELIST.extend(whitelist)
        super().__init__(app)

    async def dispatch(self, request: Request, next_func) -> Response:
        """
        接口鉴权
        :param request:
        :param next_func:
        :return:
        """
        try:
            rs = share_auth.reload(self.FASTAPI_APP.routes, WHITELIST).auth_check(request)
            if rs:
                resp: StreamingResponse = await next_func(request)
                return resp
            return JSONResponse({'success': False, 'msg': '无权访问', 'code': ErrCode.ERROR}, status_code=403)
        except NoAuthException as na_ex:
            return JSONResponse({'success': False, 'msg': na_ex.message, 'code': ErrCode.ERROR}, status_code=401)
        except Exception as ex:
            req_id = str(uuid.uuid4())
            LOGGER.error(f'[requestId={req_id}] {traceback.format_exc()}')
            return JSONResponse({'success': False, 'msg': f'服务异常[requestId={req_id}]', 'code': ErrCode.ERROR}, status_code=500)



from functools import wraps


def singleton(cls):
    _instance = {}

    @wraps(cls)
    def _singlenton(*args, **kargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kargs)
        return _instance[cls]

    return _singlenton



"""
Basic models
"""
from pydantic import BaseModel


class BaseQuery(BaseModel):
    """
    basic query with nothing
    """


class PageQuery(BaseQuery):
    """
    basic pageable query
    """
    page: int = 1
    limit: int = 20

    def setPageable(self, flag):
        if not flag:
            self.page = -1
            self.limit = -1



# encoding:utf-8
import os.path
import re
import uuid
import pandas as pd
from math import isnan
from openpyxl import Workbook
from pandas import Index
from starlette.responses import FileResponse
from ..exception.exceptions import BizException
from ..lib import cfg, logger
from ..lib import time as time_lib

LOGGER = logger.get('导出')


def export_excel(data_list, schema_list, file_id=None):
    if not file_id:
        file_id = str(uuid.uuid4())
    if not data_list:
        LOGGER.info(f'导出失败: 数据为空 file_id={file_id}')
        raise BizException('无结果数据导出')

    # 结构组装
    columns = list()
    headers = list()
    configs = list()
    for meta in schema_list:
        columns.append(meta.get('col'))
        headers.append(meta.get('title'))
        if 'type' in meta:
            configs.append(meta)

    if not len(columns) or not len(headers):
        LOGGER.info(f'导出失败: 表格列配置异常 file_id={file_id}')
        raise BizException('表格列配置异常')

    # 基础目录
    base_path = f'{cfg.get("DOWNLOAD_FILE_DIR", "/data/work/files/export")}'
    if not os.path.exists(base_path):
        os.makedirs(base_path)
    # 文件路径
    file = f'{base_path}/{file_id}.xlsx'
    # 文件写入本地(两种写入方式)
    _write_with_pandas(file, data_list, columns, headers, configs)
    # _write_with_openpyxl(file, data_list, columns, headers)
    LOGGER.info(f'导出成功: file_id={file_id}, rows={len(data_list)}, file={file}')

    # 构建导出响应
    headers = {'content-type': 'application/vnd.ms-excel', 'content-disposition': f'attachment;filename={file_id}.xlsx'}
    return FileResponse(file, media_type='xls/xlsx', headers=headers)


# 通过pandas写excel
def _write_with_pandas(file, data_list, columns, headers, configs):
    # 本地写入Excel
    write = pd.ExcelWriter(file, engine='xlsxwriter', options={'strings_to_urls': False})
    df = pd.DataFrame(data_list)

    # 验证列与数据字段是否匹配，有无缺少
    if not len(Index(columns) & df.columns) or len(Index(columns) & df.columns) != len(columns):
        # 数据中不存在的列（常见于某字段在数据集中不存在）
        miss_columns = list(set(columns) - set((Index(columns) & df.columns).tolist()))
        if miss_columns:
            for field in miss_columns:
                # 补充字段，设置默认值为空字符串
                df.insert(loc=0, column=field, value='')
        df = df.reindex(columns=columns)

    if configs:
        _add_config(df, configs)

    df.to_excel(write, sheet_name='数据', columns=columns, header=headers, index=False)
    write.save()


# 特定列数据转换
def _add_config(df, configs):
    for row in configs:
        col = row.get('col')
        type = row.get('type')
        config = row.get('excel', {})
        if type == 'timestamp':
            _process_timestamp_column(df, col, config)
        elif type == 'mapper':
            _process_mapper_column(df, col, config)


# 处理时间戳格式
def _process_timestamp_column(df, col, config={}):
    format_str = config.get('format', '%Y-%m-%d %H:%M:%S')
    df[col] = df[col].apply(lambda t: time_lib.format_timestamp10(t, fmt=format_str) if t and not isnan(t) else '')


# 处理boolean格式
def _process_mapper_column(df, col, config={}):
    bool_map = config.get('mapper', {True: '是', False: '否'})
    df[col] = df[col].apply(lambda x: bool_map.get(x) or '')


# 通过openpyxl写excel
def _write_with_openpyxl(file, data_list, columns, headers):
    wb = Workbook()
    ws = wb.active
    # 表头
    for col in range(len(headers)):
        c = col + 1
        ws.cell(row=1, column=c).value = headers[col]
    # 插入数据
    row_idx = 1
    for item in data_list:
        row = []
        for col in range(len(headers)):
            cell_val = _remove_special(item.get(columns[col], ''))
            row.append(cell_val)
        ws.append(row)
        row_idx += 1
    wb.save(filename=file)


def _remove_special(value):
    sub_str = re.sub(u'([^\u4e00-\u9fa5\u0030-\u0039\u0041-\u005a\u0061-\u007a])', '', str(value))
    return sub_str



"""
Basic models
"""
from typing import Tuple, Any
from pydantic import BaseModel

# basic function call return value, with only success flag and err msg
PCallRet = Tuple[bool, str]


class ExecRet(BaseModel):
    """
    basic ret value for executing a specific task
    """
    success: bool = True
    msg: str = ''
    data: Any = None

    @classmethod
    def ok(cls, **kwargs):
        return cls(success=True, **kwargs)

    @classmethod
    def err(cls, **kwargs):
        return cls(success=False, **kwargs)



# -*- coding: utf-8 -*-
from ..lib import cfg, logger
from ..exception.exceptions import NoConfigException

LOGGER = logger.get('CeleryClient')


class ZcbotCelery(object):

    @staticmethod
    def init(celery_broker_url: str = None, celery_result_backend: str = None, monitor_redis_uri: str = None, app_code: str = None, monitor_on: bool = True):
        """
        初始化，全局调用一次
        :param celery_broker_url:
        :param celery_result_backend:
        :param monitor_redis_uri:
        :param app_code:
        :param monitor_on:
        :return:
        """
        from zcbot_celery_sdk.client import CeleryClientHolder
        from zcbot_celery_sdk.monitor import CeleryRedisResultMonitor

        _celery_broker_url = celery_broker_url or cfg.get('CELERY_BROKER')
        _celery_result_backend = celery_result_backend or cfg.get('CELERY_BACKEND')
        _monitor_redis_uri = monitor_redis_uri or cfg.get('CELERY_MONITOR_REDIS_URI')
        _app_code = app_code or cfg.get('APP_CODE')
        if not _celery_broker_url:
            raise NoConfigException('Celery broker_url not config!')
        if not _celery_result_backend:
            raise NoConfigException('Celery result_backend not config!')
        if not _monitor_redis_uri:
            raise NoConfigException('Celery monitor_redis_uri not config!')
        if not _app_code:
            raise NoConfigException('Celery app_code not config!')

        # 全局初始化一个默认client实例
        CeleryClientHolder.init_default_instance(
            celery_broker_url=_celery_broker_url,
            celery_result_backend=_celery_result_backend,
            monitor_redis_uri=_monitor_redis_uri,
            app_code=_app_code
        )
        LOGGER.info(f'init zcbot celery client: broker={_celery_broker_url}, backend={_celery_result_backend}, monitor_redis={_monitor_redis_uri}')

        if monitor_on:
            # 结果监听器（异步调用时需要）
            _monitor = CeleryRedisResultMonitor(
                celery_broker_url=_celery_broker_url,
                celery_result_backend=_celery_result_backend,
                monitor_redis_uri=_monitor_redis_uri,
                app_code=_app_code
            )
            _monitor.start()
            LOGGER.info(f'init zcbot celery monitor: broker={_celery_broker_url}, backend={_celery_result_backend}, monitor_redis={_monitor_redis_uri}')



# -*- coding: utf-8 -*-
from ..lib import cfg, logger

LOGGER = logger.get('RedisClient')


# @singleton(不要加单例,目前项目已经不支持)
class Redis(object):
    """
    Redis客户端简易封装（单例）
    """

    def __init__(self, redis_uri=None, redis_db=None, decode_responses=True):
        import redis

        if not redis_uri:
            redis_uri: str = cfg.get('REDIS_URL') or 'redis://localhost:6379'
        if not redis_db:
            redis_db: str = cfg.get('REDIS_DB')
        self.client = redis.Redis.from_url(url=redis_uri, db=redis_db, decode_responses=decode_responses)
        LOGGER.info(f'init redis client: uri={redis_uri}, db={redis_db}')

    # 关闭链接
    def close(self):
        self.client.close()

    # 销毁关闭连接池
    def __del__(self):
        if self.client:
            try:
                self.client.close()
            except Exception as e:
                print(e)



# -*- coding: UTF-8 -*-
import json
import uuid
import hashlib
import time
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from base64 import b64encode, b64decode

chars = ["a", "b", "c", "d", "e", "f",
         "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s",
         "t", "u", "v", "w", "x", "y", "z", "0", "1", "2", "3", "4", "5",
         "6", "7", "8", "9", "A", "B", "C", "D", "E", "F", "G", "H", "I",
         "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V",
         "W", "X", "Y", "Z"]


def generate_app_id():
    """
    生成APP ID
    """
    short_buffer = []
    uuid_str = str(uuid.uuid4()).replace('-', '')

    for i in range(8):
        substr = uuid_str[i * 4: i * 4 + 4]
        x = int(substr, 16)
        short_buffer.append(chars[x % 0x3E])

    return ''.join(short_buffer)


def generate_access_key(data: dict):
    """
    生成Access Key
    """
    # 将数据转换为 UTF-8 编码的 JSON 字符串
    utf8_string = json.dumps(data).encode('utf-8')

    # 对 UTF-8 编码的 JSON 字符串进行 base64 编码
    encoded_data = b64encode(utf8_string)

    # 将 base64 编码后的数据与 app_id 进行拼接
    combined_data = encoded_data.decode("utf-8")

    # 计算拼接后字符串的 MD5 摘要
    md5_hash = hashlib.md5(combined_data.encode('utf-8')).hexdigest()

    return md5_hash


def generate_secret_key():
    """
    生成 Secret Access Key
    """
    # 生成 Secret Key，这里使用当前时间戳和一些随机性
    secret_key_data = f"{time.time()}:{uuid.uuid4()}"
    secret_key = hashlib.sha256(secret_key_data.encode()).hexdigest()

    return secret_key


# 加密
def encrypt_aes(data, key: str, iv: str):
    # 使用 AES 算法，CBC 模式进行加密
    cipher = Cipher(algorithms.AES(key.encode('utf-8')), modes.CFB(iv.encode('utf-8')), backend=default_backend())
    encryptor = cipher.encryptor()
    utf8_string = json.dumps(data).encode('utf-8').decode('utf-8')
    # 加密数据
    encrypted_data = encryptor.update(utf8_string.encode('utf-8')) + encryptor.finalize()

    # 将加密后的数据进行 Base64 编码
    return b64encode(encrypted_data).decode('utf-8')


# 解密
def decrypt_aes(encrypted_data, key: str, iv: str):
    # 使用 AES 算法，CBC 模式进行解密
    cipher = Cipher(algorithms.AES(key.encode('utf-8')), modes.CFB(iv.encode('utf-8')), backend=default_backend())
    decryptor = cipher.decryptor()
    # Base64 解码
    encrypted_data_bytes = b64decode(encrypted_data)
    # 解密数据
    decrypted_data = decryptor.update(encrypted_data_bytes) + decryptor.finalize()

    return decrypted_data.decode('utf-8')

# if __name__ == '__main__':
#     data_to_encrypt = {
#         "id": 1,
#         'userId': '1067246875800000001',
#         'username': 'admin',
#         'realName': '超级管理员',
#         'tenantCode': 'system'
#     }
#     print(generate_app_id())
#     encrypted_data = encrypt_aes(data_to_encrypt, key, iv)
#     print(f'Encrypted: {encrypted_data}')
#     #
#     decrypted_data = decrypt_aes(encrypted_data, key, iv)
#     print(f'Decrypted: {decrypted_data}')



import os
from typing import Any


def get(key: str, default: Any = '') -> Any:
    """
    get one config value
    :param key: config key
    :param default: default value
    :return: config value
    """
    return os.getenv(key.upper(), default)


def get_str(key: str) -> str:
    """
    get string config value
    :param key: config key
    :return: string config value
    """
    return str(get(key))


def get_bool(key: str) -> bool:
    """
    get bool config value
    :param key: config key
    :return: bool config value
    """
    return str(get(key)).upper() == 'TRUE'


def get_int(key: str, panic: bool = False) -> int:
    """
    get int config value
    :param key: config key
    :param panic: will raise exception if fail
    :return: int config vallue
    """
    try:
        return int(str(get(key)))
    except Exception as e:
        if panic:
            raise e
        return 0


def get_float(key: str, panic: bool = False) -> float:
    """
    get float config  value
    :param key: config key
    :param panic: will raise exception if fail
    :return: float config value
    """
    try:
        return float(str(get(key)))
    except Exception as e:
        if panic:
            raise e
        return 0.0



# -*- coding: utf-8 -*-
import pika
from ..exception.exceptions import NoConfigException
from ..lib import cfg, logger

LOGGER = logger.get('RabbitClient')


# RabbitMQ客户端简易封装
class Rabbit(object):

    def __init__(self):
        self.connection_uri = cfg.get('RABBITMQ_URI')
        if not self.connection_uri:
            raise NoConfigException('rabbitmq uri not config!')
        self.connection = pika.BlockingConnection(pika.URLParameters(self.connection_uri))
        LOGGER.info(f'init rabbit client: connection_uri -> {self.connection_uri}')

    def get_channel(self):
        return self.connection.channel()

    def reconnect(self):
        """init a new connection"""
        if self.connection:
            try:
                self.connection.close()
            except Exception as e:
                LOGGER.error(e)

        self.connection = pika.BlockingConnection(pika.URLParameters(self.connection_uri))

    def close(self):
        """close a connection"""
        if self.connection and self.connection.is_open:
            try:
                self.connection.close()
            except Exception as e:
                LOGGER.error(e)



from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware.base import BaseHTTPMiddleware
from ..auth.share_auth import ShareAuth
from ..context.context_vars import auth_user_context, tenant_context
from ..lib import logger

LOGGER = logger.get('租户中间件')
share_auth = ShareAuth()


class TenantMiddleware(BaseHTTPMiddleware):
    """ 租户中间件 """

    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request: Request, next_func) -> Response:
        """
        租户挂载中间件
        :param request:
        :param next_func:
        :return:
        """
        # 挂载
        auth_user = share_auth.get_auth_user(request)
        auth_user_context.set(auth_user)
        if auth_user and auth_user.tenantCode:
            # 复用auth_user避免重复解析用户
            tenant_context.set(auth_user.tenantCode)
        # 执行
        response = await next_func(request)
        # 清理
        tenant_context.set(None)
        auth_user_context.set(None)

        return response



# encoding:utf-8
from enum import Enum


class UserStatus(int, Enum):
    """
    batch status enum
    """
    # 正常
    NORMAL = 1
    # TODO 待确定 -> 锁定/冻结
    LOCKED = 2


class UserGender(int, Enum):
    """
    gender enum
    """
    # 男
    MALE = 1
    # 女
    FEMALE = 2
    # 保密
    SECRET = 3



import logging


def get(name: str = '', tag: str = '') -> logging.Logger:
    """
    get logger of specific tag
    :param name: name of the logger
    :param tag: tag of the logger
    :return: logger instance
    """
    if tag:
        logger_name = (tag + '_' + name).upper()
    else:
        logger_name = name.upper()
    return logging.getLogger(logger_name)


def for_handler(name: str) -> logging.Logger:
    """
    get handler logger
    :param name: handler name
    :return: controller logger
    """
    return get('handler', name)


def for_middleware(name: str) -> logging.Logger:
    """
    get middleware logger
    :param name: middleware name
    :return: middleware logger
    """
    return get('middleware', name)


def for_model(name: str) -> logging.Logger:
    """
    get model logger
    :param name: model name
    :return: model logger
    """
    return get('model', name)


def for_service(name: str) -> logging.Logger:
    """
    get service logger
    :param name: service name
    :return: service logger
    """
    return get('service', name)


def for_job(name: str) -> logging.Logger:
    """
    get job logger
    :param name: job name
    :return: job logger
    """
    return get('job', name)


def for_provider(name: str) -> logging.Logger:
    """
    get provider logger
    :param name: provider name
    :return: provider logger
    """
    return get('provider', name)


def for_util(name: str) -> logging.Logger:
    """
    get util logger
    :param name: util name
    :return: util logger
    """
    return get('util', name)


def for_client(name: str) -> logging.Logger:
    """
    get client logger
    :param name: client name
    :return: client logger
    """
    return get('client', name)



import asyncio
import functools
import typing
from concurrent.futures import ThreadPoolExecutor

T = typing.TypeVar('T')

# worker pool starts in main process
_WORKER_POOL: typing.Optional[ThreadPoolExecutor] = None

# task pool launched in workers
_TASK_POOL: typing.Optional[ThreadPoolExecutor] = None


def get_worker_pool() -> ThreadPoolExecutor:
    """
    get worker pool instance
    should be executed in main process
    main process -> worker 1 -> task 1
                                task 2
                                task 3
                    worker 2 -> task 1
                                task 2
                                task 3
                    worker 3 -> task 1
                                task 2
                                task 3
    task pool should be released once tasks are finished
    while worker pool should be always held at background
    :return: worker pool instance
    """
    global _WORKER_POOL
    if not _WORKER_POOL:
        _WORKER_POOL = ThreadPoolExecutor()
    return _WORKER_POOL


def get_task_pool() -> ThreadPoolExecutor:
    """
    get task pool instance
    should be executed in worker logics
    :return: task pool instance
    """
    global _TASK_POOL
    if not _TASK_POOL:
        _TASK_POOL = ThreadPoolExecutor()
    return _TASK_POOL


def get_task_pool_size() -> int:
    """
    get task pool instance
    should be executed in worker logics
    :return: task pool instance
    """
    global _TASK_POOL
    if _TASK_POOL:
        return _TASK_POOL._work_queue.qsize()
    return 0


def release_task_pool():
    """
    release task pool instance to avoid memory leak
    :return: None
    """
    global _TASK_POOL
    if not _TASK_POOL:
        return
    _TASK_POOL.shutdown()
    _TASK_POOL = None


async def run_in_worker(func: typing.Callable[..., T],
                        *args: typing.Any,
                        **kwargs: typing.Any) -> T:
    """
    run single task in worker
    :param func: function
    :param args: args
    :param kwargs: keyword args
    :return: result of the function
    """
    loop = asyncio.get_event_loop()
    f = functools.partial(func, **kwargs)
    worker_pool = get_worker_pool()
    return await loop.run_in_executor(worker_pool, f, *args)



# -*- coding: utf-8 -*-
import logging
import re
import json
import threading
from typing import Optional, List
from starlette.routing import BaseRoute
from fastapi.routing import APIRoute
from starlette.requests import Request
from ..lib import cfg
from ..auth import auth_cache_pool
from ..client.redis_client import Redis
from ..exception.exceptions import NoAuthException
from ..model.items import AuthUser
from ..context.context_vars import tenant_context

logger = logging.getLogger(__name__)

# TODO 暂时如此
FIX_WHITELIST = [
]


class ShareAuth(object):
    _instance_lock = threading.Lock()
    _biz_inited = False
    _routes_inited = False

    def __new__(cls, *args, **kwargs):
        if not hasattr(ShareAuth, "_instance"):
            with ShareAuth._instance_lock:
                if not hasattr(ShareAuth, "_instance"):
                    ShareAuth._instance = object.__new__(cls)
        return ShareAuth._instance

    def __init__(self):
        if not self._biz_inited:
            self._biz_inited = True
            self.rds = Redis(redis_uri=cfg.get('AUTH_REDIS_URL', None), redis_db=cfg.get('AUTH_REDIS_DB', None)).client
            # route path regex -> tags
            self.regex_to_tags_map = dict()
            # no auth path whitelist
            self.auth_whitelist = []

    def get_access_token(self, request: Request):
        """
        解析请求头鉴权信息
        :param request:
        :return:
        """
        if request and (request.headers or request.query_params):
            access_token = request.headers.get('Authorization', '').replace('Bearer ', '')
            if not access_token:
                access_token = request.cookies.get('access_token')
            if not access_token:
                access_token = request.query_params.get('access_token')
            return access_token

        return ''

    def get_api_token(self, request: Request) -> str:
        """
        api接口解析token
        """
        if request and (request.headers or request.query_params):
            token_str = request.headers.get('token')
            if not token_str:
                token_str = request.cookies.get('token')
            if not token_str:
                token_str = request.query_params.get('token')
            return token_str

        return ""

    def reload(self, routes: Optional[List[BaseRoute]], white_list=[], force: bool = False):
        if not force and self._routes_inited:
            return self
        # 构造路由匹配正则与权限颗粒(tags)映射
        _tmp_regex_to_tags_map = dict()
        for route in routes:
            if isinstance(route, APIRoute) and route.path_regex:
                _tmp_regex_to_tags_map[route.path_regex] = route.tags
        self.regex_to_tags_map.clear()
        self.regex_to_tags_map = _tmp_regex_to_tags_map
        logger.info(f'reload path to tags map: {len(self.regex_to_tags_map)}')

        # 构造权限白名单
        _auth_whitelist = list()
        for row in FIX_WHITELIST:
            _auth_whitelist.append(re.compile(row))
        for row in white_list:
            _auth_whitelist.append(re.compile(row))
        self.auth_whitelist.clear()
        self.auth_whitelist = _auth_whitelist
        logger.info(f'reload no auth whitelist: {len(self.auth_whitelist)}')
        #
        self._routes_inited = True

        return self

    def access_check(self, request: Request):
        """
        登录校验
        :param request:
        :return:
        """
        access_token = self.get_access_token(request)
        if not access_token:
            raise NoAuthException('请先登录')

        if not auth_cache_pool.has_user(access_token):
            raise NoAuthException('登录已失效，请重新登录')

        return access_token

    def auth_check(self, request: Request):
        """
        权限校验
        :param request:
        :return:
        """
        if not self.regex_to_tags_map:
            logger.debug(f'no path to tags map')
            return False
        # 优先处理白名单
        for white_regex in self.auth_whitelist:
            if white_regex.match(request.url.path):
                return True
        # 先校验登录
        access_token = self.access_check(request)
        for regex in self.regex_to_tags_map.keys():
            if regex.match(request.url.path):
                # 匹配到路由
                tags = self.regex_to_tags_map.get(regex)
                if tags:
                    # 需要登录且需要鉴权
                    auth_rs = self._rds_permissions_check(tags, access_token)
                    if not auth_rs:
                        raise NoAuthException('权限不足')
                    return True
                else:
                    # 仅登录无需鉴权
                    return True

        return False

    def _rds_permissions_check(self, tags, access_token) -> bool:
        # redis权限查询
        for tag in tags:
            has = auth_cache_pool.has_permission(access_token, json.dumps(tag))
            if not has:
                has = auth_cache_pool.has_permission(access_token, tag)
                if not has:
                    return False

        return True

    def get_auth_user(self, request: Request) -> Optional[AuthUser]:
        """
        获取当前登录用户信息
        :param request:
        :return:
        """
        access_token = self.get_access_token(request)
        if access_token and auth_cache_pool.has_user(access_token):
            auth_user = auth_cache_pool.get_cached_user(access_token)
            # 仅超级管理员可以切换租户
            if auth_user and auth_user.superAdmin == '1':
                tenant_code = request.headers.get('TenantCode', None)
                if tenant_code:
                    # 切换租户
                    auth_user.tenantCode = tenant_code

            return auth_user
        return None

    def get_auth_api(self, request: Request) -> Optional[AuthUser]:
        """
        api接口通过headers.token 查找对应的用户
        """
        token_str = self.get_access_token(request)
        if token_str and auth_cache_pool.has_user(token_str):
            auth_user = auth_cache_pool.get_user_by_token(token_str=token_str)
            if auth_user:
                return auth_user

        return None

    def get_auth_tenant(self, request: Request):
        """
        获取授权用户所属租户
        :param request:
        :return:
        """
        auth_user = self.get_auth_user(request)
        if auth_user:
            return auth_user.tenantCode
        return ''


# 用于handler来Depends，获取授权用户
async def authed_user(request: Request) -> Optional[AuthUser]:
    return ShareAuth().get_auth_user(request)


# 用于handler来Depends，获取授权用户，使用api推送数据
async def authed_api(request: Request) -> Optional[AuthUser]:
    return ShareAuth().get_auth_api(request)


# async def authed_api(request: Request) -> Optional[AuthUser]:
#     auth_user = ShareAuth().get_auth_api(request)
#     if auth_user and auth_user.tenantCode:
#
#         tenant_context.set(auth_user.tenantCode)
#     else:
#         raise NoAuthException()
#     return auth_user


# 用于handler来Depends，获取授权用户所属租户
async def authed_tenant(request: Request) -> Optional[str]:
    auth_user = ShareAuth().get_auth_user(request)
    if auth_user:
        return auth_user.tenantCode
    return None


# 用于handler来Depends，获取授权用户
async def access_token(request: Request) -> Optional[str]:
    return ShareAuth().get_access_token(request)



# -*- coding: utf-8 -*-
import pytz
from typing import Dict
from pymongo import MongoClient
from ..exception.exceptions import NoConfigException
from ..lib import cfg, logger

LOGGER = logger.get('MongoClient')


class Mongo(object):
    """
    Mongo基础客户端简易封装
    """

    def __init__(self, mongo_uri=None, mongo_db=None, *args, **kwargs):
        self.mongo_uri = mongo_uri or cfg.get('MONGO_URL')
        self.mongo_db = mongo_db or cfg.get('MONGO_DB')
        if not self.mongo_uri:
            raise NoConfigException('mongodb uri not config!')
        if not self.mongo_db:
            raise NoConfigException('mongodb database not config!')

        self.client = MongoClient(self.mongo_uri, tz_aware=True, tzinfo=pytz.timezone('Asia/Shanghai'))
        self.db = self.client[self.mongo_db]
        LOGGER.info(f'init mongo client: uri={self.mongo_uri}, db={self.mongo_db}')

    # 获取集合
    def get_collection(self, coll):
        return self.db.get_collection(coll)

    # 查询对象
    def get(self, collection, query={}):
        result = self.db[collection].find_one(query)
        return result

    # 统计数量
    def count(self, collection, query={}):
        return self.db[collection].count(query)

    # 查询列表
    def list(self, collection, query={}, fields=None, sort=[]):
        cursor = self.db[collection].find(filter=query, projection=fields, sort=sort)
        return list(cursor)

    # 分页查询
    def page(self, collection, query={}, page_no=1, page_size=20, fields=None, sort=[]):
        total = self.db[collection].count_documents(query) or 0
        cursor = self.db[collection].find(query, fields, sort=sort).skip(page_size * (page_no - 1)).limit(page_size)
        rows = list(cursor) or []

        return rows, total

    # 查询列表前N个
    def top(self, collection, query={}, sort=[], limit=1, fields=None):
        cursor = self.db[collection].find(filter=query, projection=fields, sort=sort, limit=limit)
        rows = list(cursor)
        if limit == 1 and len(rows):
            return rows[0]

        return rows

    # 查询去重列表
    def distinct(self, collection, dist_key, query={}, fields=None):
        return self.db[collection].find(query, fields).distinct(dist_key)

    # 含分页聚合查询
    def aggregate_page(self, collection, pipelines, page_no=1, page_size=20):
        skip = page_size * (page_no - 1)
        if pipelines:
            pipelines.append({'$facet': {'total': [{'$count': 'count'}], 'rows': [{'$skip': skip}, {'$limit': page_size}]}})
            pipelines.append({'$project': {'data': '$rows', 'total': {'$arrayElemAt': ['$total.count', 0]}}})

            rs = list(self.db[collection].aggregate(pipelines, session=None, allowDiskUse=True))
            if rs and 'data' in rs[0] and 'total' in rs[0]:
                return rs[0].get('data'), rs[0].get('total')

        return [], 0

    # 聚合查询
    def aggregate(self, collection, pipelines=[]):
        cursor = self.db[collection].aggregate(pipelines, session=None, allowDiskUse=True)
        return list(cursor)

    # 查询分页列表
    def list_with_page(self, collection, query={}, page_size=10000, fields=None):
        rows = list()
        total = self.db[collection].count(query)
        if total > 0 and page_size > 0:
            total_page = round(len(total) / page_size)
            for page in range(0, total_page):
                if fields:
                    cursor = self.db[collection].find(query, fields).skip(page_size * page).limit(page)
                else:
                    cursor = self.db[collection].find(query).skip(page_size * page).limit(page)
                curr_batch = list(cursor)
                if curr_batch:
                    rows.append(curr_batch)
        return rows

    # 插入或更新
    def insert_or_update(self, collection, data, id_key='_id', upsert=True, multi=False):
        return self.db[collection].update({id_key: data[id_key]}, {'$set': data}, upsert=upsert, multi=multi)

    # 插入或更新
    def insert(self, collection, data):
        return self.db[collection].insert_one(data)

    # 更新
    def update(self, collection, filter, data, multi=False):
        return self.db[collection].update(filter, {'$set': data}, multi=multi)

    # 原生保存方法
    def save(self, collection, filter, save_data: Dict, upsert=True):
        return self.db[collection].update(filter, save_data, upsert=upsert)

    # 以主键更新
    def update_by_pk(self, collection, pk_val, data, upsert=False):
        return self.db[collection].update({'_id': pk_val}, {'$set': data}, multi=False, upsert=upsert)

    # 批量更新
    def batch_update(self, collection, filter, datas, multi=True):
        return self.db[collection].update(filter, datas, multi=multi)

    # 更新
    def delete(self, collection, filter):
        return self.db[collection].delete_many(filter)

    # 插入或更新
    def bulk_write(self, collection, bulk_list):
        if bulk_list:
            return self.db[collection].bulk_write(bulk_list, ordered=False, bypass_document_validation=True)

    # 创建索引
    def create_index(self, collection, fields):
        return self.db[collection].create_index(fields)

    # 关闭链接
    def close(self):
        self.client.close()

    # 销毁关闭链接
    def __del__(self):
        if self.client:
            try:
                self.client.close()
            except Exception as e:
                print(e)
